<?php

namespace Drupal\phone_verification\Plugin\Field\FieldWidget;

use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\phone_verification\PhoneVerificationInterface;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\telephone\Plugin\Field\FieldWidget\TelephoneDefaultWidget;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'phone_verification_telephone' widget.
 *
 * @FieldWidget(
 *   id = "phone_verification_telephone",
 *   label = @Translation("Phone verification telephone number"),
 *   field_types = {
 *     "telephone"
 *   }
 * )
 */
class PhoneVerificationTelephoneWidget extends TelephoneDefaultWidget implements ContainerFactoryPluginInterface {

  protected $config;

  protected $phone_verification;

  protected $module_handler;

  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, array $third_party_settings, PhoneVerificationInterface $phone_verification, ModuleHandlerInterface $module_handler) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings);
    $this->config = \Drupal::config('phone_verification.settings');
    $this->phone_verification = $phone_verification;
    $this->module_handler = $module_handler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['third_party_settings'],
      $container->get('phone_verification'),
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    $settings =  parent::defaultSettings();

    $settings['code_type'] = NULL;
    $settings['code_length'] = NULL;
    $settings['sms_message'] = NULL;

    return $settings;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary =  parent::settingsSummary();

    $code_length = $this->getSetting('code_length');
    if ($code_length) {
      $summary[] = $this->t('Verification code length: @length',['@length' => $code_length]);
    }

    $code_type = $this->getSetting('code_type');
    if (isset($code_types[$code_type])) {
      $code_types = $this->phone_verification->getCodeTypes();
      $summary[] = $this->t('Verification code type: @type',['@type' => $code_types[$code_type]]);
    }

    $sms_message = $this->getSetting('sms_message');
    if ($sms_message !== NULL) {
      $summary[] = $this->t('SMS Message: @sms',['@sms' => $sms_message]);
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element = parent::settingsForm($form, $form_state);

    $element['notice'] = [
      '#markup' => $this->t('Override global module settings here. If you leave a setting empty, global value will be used.')
    ];

    $element['code_length'] = [
      '#type' => 'number',
      '#title' => $this->t('Verification code length'),
      '#default_value' => $this->getSetting('code_length'),
    ];

    $element['code_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Verification code type'),
      '#default_value' => $this->getSetting('code_type'),
      '#empty_option' => $this->t('Use default'),
      '#options' => $this->phone_verification->getCodeTypes(),
    ];

    $element['sms_message'] = [
      '#type' => 'textarea',
      '#title' => $this->t('SMS Message'),
      '#value' => $this->getSetting('sms_message'),
      '#description' => $this->t('SMS Message must contain @code placeholder for generated code. You can use core tokens in this field.
        To use advanced tokens and token browser, install token module.'),
      '#default_value' => $this->t('@code'),
      '#element_validate' => [[$this,'smsMessageFieldValidator']],
    ];

    if ($this->module_handler->moduleExists('token')) {
      $element['token_browser'] = [
        '#theme' => 'token_tree_link',
        '#click_insert' => FALSE,
        '#dialog' => TRUE,
      ];
    }

    return $element;
  }

  /**
   * Validation function for message field.
   */
  public function smsMessageFieldValidator($element, FormStateInterface $form_state) {
    $value = $form_state->getValue($element['#parents']);

    // Check if SMS message contains @code placeholder.
    if ($value && strpos($value,'@code') === FALSE) {
      $form_state->setError($element,$this->t('SMS Message must contain @code placeholder for generated code'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    $element = parent::formElement($items, $delta, $element, $form, $form_state);

    $element['#element_validate'][] = [$this, 'elementValidate'];

    $element['#process'][] = [get_class($this), 'process'];

    $field_name = $this->fieldDefinition->getName();
    $wrapper_id = Html::getUniqueId($field_name . '-phone-verification-wrapper');

    $element['value']['#weight'] = -1;
    $element['value']['#ajax'] = [
      'callback' => [$this,'updateElementCallback'],
      'event' => 'change',
      'wrapper' => $wrapper_id
    ];

    $element['#prefix'] = '<div id="' . $wrapper_id . '">';
    $element['#suffix'] = '</div>';

    $value = $form_state->getValue(array_merge($element['#field_parents'], [$field_name, $delta]));

    $number = NULL;
    if (!empty($value['value'])) {
      $number = $value['value'];
    }
    elseif (!empty($element['value']['#default_value'])) {
      $number = $element['value']['#default_value'];
    }
    $show_phone_verification = !empty($number);
    $verified = $this->phone_verification->isVerifiedNumber($number);

    $show_verify = !empty($value['phone_verification']['show_verify']);

    if ($show_phone_verification && !$verified) {
      $element['phone_verification']['send_sms'] = [
        '#name' => 'phone_verification_send_sms',
        '#type' => 'submit',
        '#submit' => [[$this,'sendSMSSubmit']],
        '#value' => $this->t('Send verification SMS'),
        '#validate' => [[$this,'sendSMSValidator']],
        '#ajax' => [
          'callback' => [$this,'updateElementCallback'],
          'wrapper' => $wrapper_id
        ],
        '#attributes' => [
          'class' => ['send-sms'],
        ],
        '#weight' => 100,
      ];
      if ($show_verify) {
        $element['phone_verification']['send_sms']['#value'] = $this->t('Resend verification SMS');
        $element['phone_verification']['sms_code'] = [
          '#type' => 'textfield',
          '#title' => $this->t('SMS verification code'),
          '#required' => TRUE,
          '#wrapper_attributes' => [
            'class' => ['sms-verification-code'],
          ],
        ];
        $element['phone_verification']['verify_number'] = [
          '#name' => 'phone_verification_verify_number',
          '#type' => 'submit',
          '#value' => $this->t('Verify'),
          '#submit' => [[$this,'verifySMSSubmit']],
          '#validate' => [[$this,'verifySMSValidator']],
          '#ajax' => [
            'callback' => [$this,'updateElementCallback'],
            'wrapper' => $wrapper_id,
          ],
          '#attributes' => [
            'class' => ['verify-number'],
          ],
        ];
      }
    }else{
      if ($verified) {
        $element['message'] = [
          '#type' => 'item',
          '#markup' => $this->t('Number verified!')
        ];
      }
      unset($element['phone_verification']);
    }

    return $element;
  }

  /**
   * Process callback.
   *
   * Limit validation errors for action buttons.
   */
  public static function process($element, FormStateInterface $form_state, $form) {
    $element['phone_verification']['send_sms']['#limit_validation_errors'] = [array_merge($element['#parents'], ['value'])];
    $element['phone_verification']['verify_number']['#limit_validation_errors'] = [$element['#parents']];
    return $element;
  }

  /**
   * Validation function for send sms button.
   *
   * Sends the message or sets form error if message could not be set.
   */
  public function sendSMSValidator(&$form, FormStateInterface &$form_state) {
    $parent_value = $this->getParentElementValue($form_state);

    // Generate code, get phone number and send SMS
    $code = $this->phone_verification->generateCode($this->getSetting('code_length'), $this->getSetting('code_type'));
    $telephone_number = $parent_value['value'];

    $message = $this->getSetting('sms_message');
    if (empty($message)) {
      $message = $this->config->get('sms_message');
    }

    $result = $this->phone_verification->sendVerificationCode($telephone_number, $code, $message);

    $parent_element = $this->getParentElement($form, $form_state);

    if (!$result || $result->getError()) {
      $form_state->setError(
        $parent_element['value'],
        $this->t('There was a problem sending an SMS, check entered phone number')
      );
    }
  }

  /**
   * Submit function for send sms button.
   *
   * Sets helper values for the element.
   */
  public function sendSMSSubmit(array &$form, FormStateInterface $form_state) {

    $parent_element = $this->getParentElement($form, $form_state);
    $form_state->setValue(array_merge($parent_element['#parents'], ['phone_verification', 'show_verify']),TRUE);
    $form_state->setRebuild(TRUE);

  }

  /**
   * Validation function for verify button.
   *
   * Verifies the validity of entered code.
   */
  public function verifySMSValidator(&$form, FormStateInterface &$form_state) {
    $parent_value = $this->getParentElementValue($form_state);

    $code = $parent_value['phone_verification']['sms_code'];
    $telephone_number = $parent_value['value'];

    $parent_element = $this->getParentElement($form, $form_state);

    $code_valid = $this->phone_verification->verifyCode($telephone_number, $code);
    if (!$code_valid) {
      $form_state->setError($parent_element['phone_verification']['sms_code'],
        t('Incorrect code or entry not valid.'));
      $form_state->setRebuild();
    }
  }

  /**
   * Submit function for verify button.
   *
   * Sets helper values for the element.
   */
  public function verifySMSSubmit(array &$form, FormStateInterface $form_state) {
    $form_state->setRebuild(TRUE);
  }

  /**
   * Ajax update callback for the element.
   */
  public function updateElementCallback(array &$form, FormStateInterface $form_state) {
    // Rebuild phone number verification form part
    return $this->getParentElement($form, $form_state);
  }

  /**
   * Validate the element on form submit
   */
  public function elementValidate($element, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    $field_name = $this->fieldDefinition->getName();
    $value = $form_state->getValue($element['#parents']);

    // Do not set errors when validation is triggered by this element or its children
    // Only trigger this error when submitting the form
    if (isset($value['phone_verification']) && !in_array($field_name, $triggering_element['#array_parents'])) {
      $form_state->setError($element['value'], $this->t('Telephone number not validated'));
    }
  }

  /**
   * Helper function to get value of the main parent element
   *
   * @return array
   */
  public function getParentElementValue(FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $triggering_element = $form_state->getTriggeringElement();
    $field_name = $this->fieldDefinition->getName();
    $field_name_index = array_search($field_name, $triggering_element['#parents']);
    return NestedArray::getValue($values, array_slice($triggering_element['#parents'], 0, $field_name_index + 2));
  }

  /**
   * Helper function to get the main parent element
   *
   * @return array
   */
  public function getParentElement(array &$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    $field_name = $this->fieldDefinition->getName();
    $field_name_index = array_search($field_name, $triggering_element['#array_parents']);
    return NestedArray::getValue($form, array_slice($triggering_element['#array_parents'], 0, $field_name_index + 3));
  }

}
